@@ -1,4 +1,6 @@
-use std::fmt;
+use core::fmt;
+use core::str;
+#[cfg(feature = "std")]
 use std::error;
 
 use TokenType;
@@ -51,6 +53,7 @@ impl fmt::Display for Error {
     }
 }
 
+#[cfg(feature = "std")]
 impl error::Error for Error {
     fn description(&self) -> &str {
         "an XML parsing error"
@@ -70,6 +73,11 @@ pub enum StreamError {
     /// An invalid name.
     InvalidName,
 
+    /// A non-XML character has occurred.
+    ///
+    /// Valid characters are: <https://www.w3.org/TR/xml/#char32>
+    NonXmlChar(char, TextPos),
+
     /// An invalid/unexpected character.
     ///
     /// The first byte is an actual one, the second one is expected.
@@ -83,25 +91,28 @@ pub enum StreamError {
     InvalidCharMultiple(u8, &'static [u8], TextPos),
 
     /// An unexpected character instead of `"` or `'`.
-    InvalidQuote(char, TextPos),
+    InvalidQuote(u8, TextPos),
 
     /// An unexpected character instead of an XML space.
     ///
     /// Includes: `' ' \n \r \t &#x20; &#x9; &#xD; &#xA;`.
-    InvalidSpace(char, TextPos),
+    InvalidSpace(u8, TextPos),
 
     /// An unexpected string.
     ///
-    /// The first string is an actual one, others - expected.
-    ///
-    /// We are using a single value to reduce the struct size.
-    InvalidString(Vec<String>, TextPos),
+    /// Contains what string was expected.
+    InvalidString(&'static str, TextPos),
 
     /// An invalid reference.
     InvalidReference,
 
     /// An invalid ExternalID in the DTD.
     InvalidExternalID,
+
+    /// A Character Data node contains an invalid data.
+    ///
+    /// Currently, only `]]>` is not allowed.
+    InvalidCharacterData,
 }
 
 impl fmt::Display for StreamError {
@@ -113,27 +124,33 @@ impl fmt::Display for StreamError {
             StreamError::InvalidName => {
                 write!(f, "invalid name token")
             }
+            StreamError::NonXmlChar(c, pos) => {
+                write!(f, "a non-XML character {:?} found at {}", c, pos)
+            }
             StreamError::InvalidChar(actual, expected, pos) => {
                 write!(f, "expected '{}' not '{}' at {}",
                        expected as char, actual as char, pos)
             }
             StreamError::InvalidCharMultiple(actual, ref expected, pos) => {
-                // &[u8] -> Vec<String>
-                let list: Vec<_> =
-                    expected.iter().map(|c| String::from_utf8(vec![*c]).unwrap()).collect();
+                let mut expected_iter = expected.iter().peekable();
 
-                write!(f, "expected '{}' not '{}' at {}",
-                       list.join("', '"), actual as char, pos)
+                write!(f, "expected ")?;
+                while let Some(&c) = expected_iter.next() {
+                    write!(f, "'{}'", c as char)?;
+                    if expected_iter.peek().is_some() {
+                        write!(f, ", ")?;
+                    }
+                }
+                write!(f, " not '{}' at {}", actual as char, pos)
             }
             StreamError::InvalidQuote(c, pos) => {
-                write!(f, "expected quote mark not '{}' at {}", c, pos)
+                write!(f, "expected quote mark not '{}' at {}", c as char, pos)
             }
             StreamError::InvalidSpace(c, pos) => {
-                write!(f, "expected space not '{}' at {}", c, pos)
+                write!(f, "expected space not '{}' at {}", c as char, pos)
             }
-            StreamError::InvalidString(ref strings, pos) => {
-                write!(f, "expected '{}' not '{}' at {}",
-                       strings[1..].join("', '"), strings[0], pos)
+            StreamError::InvalidString(expected, pos) => {
+                write!(f, "expected '{}' at {}", expected, pos)
             }
             StreamError::InvalidReference => {
                 write!(f, "invalid reference")
@@ -141,10 +158,14 @@ impl fmt::Display for StreamError {
             StreamError::InvalidExternalID => {
                 write!(f, "invalid ExternalID")
             }
+            StreamError::InvalidCharacterData => {
+                write!(f, "']]>' is not allowed inside a character data")
+            }
         }
     }
 }
 
+#[cfg(feature = "std")]
 impl error::Error for StreamError {
     fn description(&self) -> &str {
         "an XML stream parsing error"
