@@ -25,6 +25,7 @@ If you are looking for a more high-level solution - checkout
 - No heap allocations.
 - No dependencies.
 - Tiny. ~1500 LOC and ~40KiB in the release build according to the `cargo-bloat`.
+- Supports `no_std` builds. To use without the standard library, disable the default features.
 
 ## Limitations
 
@@ -44,16 +45,20 @@ If you are looking for a more high-level solution - checkout
 - The library forbids the unsafe code.
 */
 
-#![cfg_attr(feature = "cargo-clippy", allow(unreadable_literal))]
+#![no_std]
 
-#![doc(html_root_url = "https://docs.rs/xmlparser/0.10.0")]
+#![doc(html_root_url = "https://docs.rs/xmlparser/0.12.0")]
 
 #![forbid(unsafe_code)]
 #![warn(missing_docs)]
 #![allow(ellipsis_inclusive_range_patterns)]
 
+#[cfg(feature = "std")]
+#[macro_use]
+extern crate std;
 
-use std::fmt;
+
+use core::fmt;
 
 
 macro_rules! matches {
@@ -237,6 +242,7 @@ pub enum Token<'a> {
     ///
     /// Contains text between elements including whitespaces.
     /// Basically everything between `>` and `<`.
+    /// Except `]]>`, which is not allowed and will lead to an error.
     ///
     /// ```text
     /// <p> text </p>
@@ -292,8 +298,8 @@ pub enum EntityDefinition<'a> {
 }
 
 
-type Result<T> = ::std::result::Result<T, Error>;
-type StreamResult<T> = ::std::result::Result<T, StreamError>;
+type Result<T> = core::result::Result<T, Error>;
+type StreamResult<T> = core::result::Result<T, StreamError>;
 
 
 /// List of token types.
@@ -337,7 +343,7 @@ impl fmt::Display for TokenType {
             TokenType::Attribute => "Attribute",
             TokenType::CDSect => "CDATA",
             TokenType::Whitespace => "Whitespace",
-            TokenType::CharData => "Character data",
+            TokenType::CharData => "Character Data",
             TokenType::Unknown => "Unknown",
         };
 
@@ -696,17 +702,32 @@ impl<'a> Tokenizer<'a> {
 
     // XMLDecl ::= '<?xml' VersionInfo EncodingDecl? SDDecl? S? '?>'
     fn parse_declaration_impl(s: &mut Stream<'a>) -> StreamResult<Token<'a>> {
+        fn consume_spaces(s: &mut Stream) -> StreamResult<()> {
+            if s.starts_with_space() {
+                s.skip_spaces();
+            } else if !s.starts_with(b"?>") && !s.at_end() {
+                return Err(StreamError::InvalidSpace(s.curr_byte_unchecked(), s.gen_text_pos()));
+            }
+
+            Ok(())
+        }
+
         let start = s.pos() - 6;
 
         let version = Self::parse_version_info(s)?;
+        consume_spaces(s)?;
+
         let encoding = Self::parse_encoding_decl(s)?;
+        if encoding.is_some() {
+            consume_spaces(s)?;
+        }
+
         let standalone = Self::parse_standalone(s)?;
 
         s.skip_spaces();
         s.skip_string(b"?>")?;
 
         let span = s.slice_back(start);
-
         Ok(Token::Declaration { version, encoding, standalone, span })
     }
 
@@ -728,16 +749,14 @@ impl<'a> Tokenizer<'a> {
         Ok(ver)
     }
 
-
     // EncodingDecl ::= S 'encoding' Eq ('"' EncName '"' | "'" EncName "'" )
     // EncName      ::= [A-Za-z] ([A-Za-z0-9._] | '-')*
     fn parse_encoding_decl(s: &mut Stream<'a>) -> StreamResult<Option<StrSpan<'a>>> {
-        s.skip_spaces();
-
-        if s.skip_string(b"encoding").is_err() {
+        if !s.starts_with(b"encoding") {
             return Ok(None);
         }
 
+        s.advance(8);
         s.consume_eq()?;
         let quote = s.consume_quote()?;
         // [A-Za-z] ([A-Za-z0-9._] | '-')*
@@ -756,12 +775,11 @@ impl<'a> Tokenizer<'a> {
 
     // SDDecl ::= S 'standalone' Eq (("'" ('yes' | 'no') "'") | ('"' ('yes' | 'no') '"'))
     fn parse_standalone(s: &mut Stream<'a>) -> StreamResult<Option<bool>> {
-        s.skip_spaces();
-
-        if s.skip_string(b"standalone").is_err() {
+        if !s.starts_with(b"standalone") {
             return Ok(None);
         }
 
+        s.advance(10);
         s.consume_eq()?;
         let quote = s.consume_quote()?;
 
@@ -772,9 +790,9 @@ impl<'a> Tokenizer<'a> {
             "yes" => true,
             "no" => false,
             _ => {
-                let values = vec![value.into(), "yes".into(), "no".into()];
                 let pos = s.gen_text_pos_from(start);
-                return Err(StreamError::InvalidString(values, pos));
+
+                return Err(StreamError::InvalidString("yes', 'no", pos));
             }
         };
 
@@ -792,18 +810,10 @@ impl<'a> Tokenizer<'a> {
     // '<!--' ((Char - '-') | ('-' (Char - '-')))* '-->'
     fn parse_comment_impl(s: &mut Stream<'a>) -> StreamResult<Token<'a>> {
         let start = s.pos() - 4;
-
-        let text = s.consume_chars(|s, c| {
-            if c == '-' && s.starts_with(b"-->") {
-                return false;
-            }
-
-            c.is_xml_char()
-        });
-
+        let text = s.consume_chars(|s, c| !(c == '-' && s.starts_with(b"-->")))?;
         s.skip_string(b"-->")?;
 
-        if text.as_str().contains("--") {
+        if text.as_str().contains("--") || text.as_str().ends_with('-') {
             return Err(StreamError::UnexpectedEndOfStream); // Error type doesn't matter.
         }
 
@@ -820,23 +830,9 @@ impl<'a> Tokenizer<'a> {
     // PITarget ::= Name - (('X' | 'x') ('M' | 'm') ('L' | 'l'))
     fn parse_pi_impl(s: &mut Stream<'a>) -> StreamResult<Token<'a>> {
         let start = s.pos() - 2;
-
         let target = s.consume_name()?;
-
         s.skip_spaces();
-
-        let content = s.consume_chars(|s, c| {
-            if c == '?' && s.starts_with(b"?>") {
-                return false;
-            }
-
-            if !c.is_xml_char() {
-                return false;
-            }
-
-            true
-        });
-
+        let content = s.consume_chars(|s, c| !(c == '?' && s.starts_with(b"?>")))?;
         let content = if !content.is_empty() {
             Some(content)
         } else {
@@ -1000,15 +996,9 @@ impl<'a> Tokenizer<'a> {
     // CDEnd   ::= ']]>'
     fn parse_cdata_impl(s: &mut Stream<'a>) -> StreamResult<Token<'a>> {
         let start = s.pos() - 9;
-
-        let text = s.consume_bytes(|s, c| {
-            !(c == b']' && s.starts_with(b"]]>"))
-        });
-
+        let text = s.consume_chars(|s, c| !(c == ']' && s.starts_with(b"]]>")))?;
         s.skip_string(b"]]>")?;
-
         let span = s.slice_back(start);
-
         Ok(Token::Cdata { text, span })
     }
 
@@ -1044,6 +1034,8 @@ impl<'a> Tokenizer<'a> {
 
     // Name Eq AttValue
     fn parse_attribute(s: &mut Stream<'a>) -> StreamResult<Token<'a>> {
+        let attr_start = s.pos();
+        let has_space = s.starts_with_space();
         s.skip_spaces();
 
         if let Ok(c) = s.curr_byte() {
@@ -1065,23 +1057,47 @@ impl<'a> Tokenizer<'a> {
             }
         }
 
+        if !has_space {
+            if !s.at_end() {
+                return Err(StreamError::InvalidSpace(
+                    s.curr_byte_unchecked(), s.gen_text_pos_from(attr_start))
+                );
+            } else {
+                return Err(StreamError::UnexpectedEndOfStream);
+            }
+        }
+
         let start = s.pos();
 
         let (prefix, local) = s.consume_qname()?;
         s.consume_eq()?;
         let quote = s.consume_quote()?;
+        let quote_c = quote as char;
         // The attribute value must not contain the < character.
-        let value = s.consume_bytes(|_, c| c != quote && c != b'<');
+        let value = s.consume_chars(|_, c| c != quote_c && c != '<')?;
         s.consume_byte(quote)?;
         let span = s.slice_back(start);
 
-        s.skip_spaces();
-
         Ok(Token::Attribute { prefix, local, value, span })
     }
 
     fn parse_text(s: &mut Stream<'a>) -> Result<Token<'a>> {
-        let text = s.consume_bytes(|_, c| c != b'<');
+        map_err_at!(Self::parse_text_impl(s), TokenType::CharData, s, 0)
+    }
+
+    fn parse_text_impl(s: &mut Stream<'a>) -> StreamResult<Token<'a>> {
+        let text = s.consume_chars(|_, c| c != '<')?;
+
+        // According to the spec, `]]>` must not appear inside a Text node.
+        // https://www.w3.org/TR/xml/#syntax
+        //
+        // Search for `>` first, since it's a bit faster than looking for `]]>`.
+        if text.as_str().contains('>') {
+            if text.as_str().contains("]]>") {
+                return Err(StreamError::InvalidCharacterData);
+            }
+        }
+
         Ok(Token::Text { text })
     }
 }
