@@ -1,6 +1,6 @@
-use std::char;
-use std::cmp;
-use std::str;
+use core::char;
+use core::cmp;
+use core::str;
 
 use {
     StreamError,
@@ -10,7 +10,7 @@ use {
     XmlCharExt,
 };
 
-type Result<T> = ::std::result::Result<T, StreamError>;
+type Result<T> = ::core::result::Result<T, StreamError>;
 
 
 /// Representation of the [Reference](https://www.w3.org/TR/xml/#NT-Reference) value.
@@ -215,20 +215,14 @@ impl<'a> Stream<'a> {
     /// # Errors
     ///
     /// - `InvalidString`
-    pub fn skip_string(&mut self, text: &[u8]) -> Result<()> {
+    pub fn skip_string(&mut self, text: &'static [u8]) -> Result<()> {
         if !self.starts_with(text) {
-            let len = cmp::min(text.len(), self.end - self.pos);
-            // Collect chars and do not slice a string,
-            // because the `len` can be on the char boundary.
-            // Which lead to a panic.
-            let actual = self.span.as_str()[self.pos..].chars().take(len).collect();
-
-            // Assume that all input `text` are valid UTF-8 strings, so unwrap is safe.
-            let expected = str::from_utf8(text).unwrap().to_owned();
-
             let pos = self.gen_text_pos();
 
-            return Err(StreamError::InvalidString(vec![actual, expected], pos));
+            // Assume that all input `text` are valid UTF-8 strings, so unwrap is safe.
+            let expected = str::from_utf8(text).unwrap();
+
+            return Err(StreamError::InvalidString(expected, pos));
         }
 
         self.advance(text.len());
@@ -259,25 +253,29 @@ impl<'a> Stream<'a> {
     /// Consumes chars by the predicate and returns them.
     ///
     /// The result can be empty.
-    pub fn consume_chars<F>(&mut self, f: F) -> StrSpan<'a>
+    pub fn consume_chars<F>(&mut self, f: F) -> Result<StrSpan<'a>>
         where F: Fn(&Stream, char) -> bool
     {
         let start = self.pos;
-        self.skip_chars(f);
-        self.slice_back(start)
+        self.skip_chars(f)?;
+        Ok(self.slice_back(start))
     }
 
     /// Skips chars by the predicate.
-    pub fn skip_chars<F>(&mut self, f: F)
+    pub fn skip_chars<F>(&mut self, f: F) -> Result<()>
         where F: Fn(&Stream, char) -> bool
     {
         for c in self.chars() {
-            if f(self, c) {
+            if !c.is_xml_char() {
+                return Err(StreamError::NonXmlChar(c, self.gen_text_pos()));
+            } else if f(self, c) {
                 self.advance(c.len_utf8());
             } else {
                 break;
             }
         }
+
+        Ok(())
     }
 
     #[inline]
@@ -384,9 +382,7 @@ impl<'a> Stream<'a> {
         }
 
         if !self.starts_with_space() {
-            let c = self.curr_byte_unchecked() as char;
-            let pos = self.gen_text_pos();
-            return Err(StreamError::InvalidSpace(c, pos));
+            return Err(StreamError::InvalidSpace(self.curr_byte_unchecked(), self.gen_text_pos()));
         }
 
         self.skip_spaces();
@@ -611,7 +607,7 @@ impl<'a> Stream<'a> {
             self.advance(1);
             Ok(c)
         } else {
-            Err(StreamError::InvalidQuote(c as char, self.gen_text_pos()))
+            Err(StreamError::InvalidQuote(c, self.gen_text_pos()))
         }
     }
 }
