@@ -45,3 +45,22 @@ test!(text_06, "<p>&#x20;</p>",
     Token::Text("&#x20;", 3..9),
     Token::ElementEnd(ElementEnd::Close("", "p"), 9..13)
 );
+
+test!(text_07, "<p>]></p>",
+    Token::ElementStart("", "p", 0..2),
+    Token::ElementEnd(ElementEnd::Open, 2..3),
+    Token::Text("]>", 3..5),
+    Token::ElementEnd(ElementEnd::Close("", "p"), 5..9)
+);
+
+test!(text_err_01, "<p>]]></p>",
+    Token::ElementStart("", "p", 0..2),
+    Token::ElementEnd(ElementEnd::Open, 2..3),
+    Token::Error("invalid token 'Character Data' at 1:4 cause ']]>' is not allowed inside a character data".to_string())
+);
+
+test!(text_err_02, "<p>\u{0c}</p>",
+    Token::ElementStart("", "p", 0..2),
+    Token::ElementEnd(ElementEnd::Open, 2..3),
+    Token::Error("invalid token 'Character Data' at 1:4 cause a non-XML character '\\u{c}' found at 1:4".to_string())
+);
